# -*- coding: utf8 -*-
from unittest import TestCase

from wordstats.language_info import LanguageInfo
from wordstats.loading_from_hermit import load_language_from_hermit, path_of_hermit_language_file
from wordstats.word_info import UnknownWordInfo
from wordstats.word_stats import Word


class SimpleTests(TestCase):
    @classmethod
    def test_loading_from_hermit_file(self):
        # see the source of load_from_hermit for details
        german = load_language_from_hermit("de")
        word = german.get("wunderbar")
        assert word.difficulty == 0.02
        assert word.klevel == 2

    @classmethod
    def test_caching_to_db(self):
        german = LanguageInfo.load_from_file(path_of_hermit_language_file("de"), "de")
        german.cache_to_db()

        # now we should be able to load from db
        deutsch = LanguageInfo.load_from_db("de")
        mutter = deutsch.get("Mutter")
        assert mutter.difficulty == 0.0
        assert mutter.klevel == 1

    @classmethod
    def test_word_stats(cls):
        assert Word.stats("spar", "de").difficulty > Word.stats("Mutter", "de").difficulty

    def test_inexistant_word(cls):
        sparalicious_info = Word.stats("sparalicious", "de")
        assert isinstance(sparalicious_info, UnknownWordInfo)
        assert sparalicious_info.importance == 0



# By importing the base service, we create the engine, and the session
from .base_service import BaseService, Base

# By importing here all the model classes, we make sure that they get
# acquinted with the Base because later the Base needs to know about
# them when it does the reflection thing to create the DB tables
from .word_info import WordInfo
from .language_info import LanguageInfo

# Create all tables in the engine. equivalent to "Create Table" in SQL
Base.metadata.create_all(BaseService.engine)

from .word_stats import Word


import codecs
import os
from datetime import datetime

from sqlalchemy import Table

from .word_info import WordInfo, UnknownWordInfo
from .utils.mem_footprint import total_size
from .base_service import BaseService, Base
from .config import MAX_WORDS
from .metrics_computers import *
import logging as log


class LanguageInfo(object):
    def __init__(self, language_id):
        self.language_id = language_id
        self.word_info_dict = dict()

    @classmethod
    def load(cls, language_code):
        from wordstats.loading_from_hermit import load_language_from_hermit

        log.info(f"loading {language_code} from DB")
        lang = LanguageInfo.load_from_db(language_code)

        if len(lang.all_words()) == 0:
            log.info(f"loading {language_code} from file")
            lang = load_language_from_hermit(language_code)
            log.info(f"caching {language_code} to DB")
            lang.cache_to_db()

        return lang

    def get(self, word):
        """

            Main use of this class is this method which retrieves
            information about a given word.

            If the word is not found we return an UnknownWordInfo
            which behaves like a WordInfo with default values.

        :param word: utf-8 string
        :return:
        """

        # TODO: think whether we want to care about the lowercase or not. For now, we don't
        word = word.lower()
        if word in self.word_info_dict:
            return self.word_info_dict[word]
        else:
            return UnknownWordInfo()

    def all_words(self):
        return list(self.word_info_dict.keys())

    def __getitem__(self, key):
        return self.get(key)

    @classmethod
    def load_from_file(cls, file_name, lang_code):
        """
        Loads the word info from the given file
        where the format expected is a word and the number
        of occurrences per line. e.g.

        der 123123
        die 123000
        das 121023
        ...

        :param file_name:
        :param lang_code:
        :return:
        """

        new_registry = cls(lang_code)

        word_rank = 0

        package_directory = os.path.dirname(os.path.abspath(__file__))

        with codecs.open(package_directory + os.sep + file_name, encoding="utf8") as words_file:
            words_list = words_file.read().splitlines()

            for word_and_freq in words_list:
                word_and_freq_array = word_and_freq.split(" ")
                word = word_and_freq_array[0]
                occurrences = int(word_and_freq_array[1])

                frequency = compute_frequency(occurrences)
                difficulty = compute_difficulty(word_rank)
                importance = compute_importance(occurrences)
                klevel = compute_klevel(word_rank)

                word_rank += 1

                if word_rank <= MAX_WORDS:

                    if word.lower() not in new_registry.word_info_dict:
                        r = WordInfo(
                            word.lower(),
                            lang_code,
                            frequency,
                            difficulty,
                            importance,
                            word_rank,
                            klevel)
                        new_registry.word_info_dict[word.lower()] = r

        return new_registry

    @classmethod
    def load_from_db(cls, language_id):
        """
        Assumes the ~./word_info file contains
        information about how to connect to the
        database. Also, assumes the database contains
        the information.

        :param language_id:
        :return:
        """

        new_registry = cls(language_id)

        all_word_info_items = WordInfo.find_all(language_id)

        for each in all_word_info_items:
            new_registry.word_info_dict[each.word] = each

        return new_registry

    def cache_to_db(self):
        """
        Useful to save the informations about a given language
        in the DB. Loading from the DB is faster than from file.
        See the tests file
        :return:
        """

        def clear_corresponding_entries_in_db(self):
            table = Table('word_info', Base.metadata, autoload=True, autoload_with=BaseService.engine)
            words = BaseService.session.query(table).filter(table.c.language_id == self.language_id)
            words.delete(synchronize_session=False)

        clear_corresponding_entries_in_db(self)
        for word_info in list(self.word_info_dict.values()):
            BaseService.session.add(word_info)
        BaseService.session.commit()

    # Everything that follows is private
    def print_load_stats(self, a, b):
        memory_footprint = total_size(self.word_info_dict, set()) / 1024 / 1024
        print(("Elapsed time to load the {0} data: {1} ({2} entries)".format(self.language_id, b - a,
                                                                             len(self.word_info_dict))))
        print(("Required memory for the {0} registry: {1}MB".format(self.language_id, memory_footprint)))

    @classmethod
    def profile_load_from_db(cls, language_id, output=False):

        a = datetime.now()
        new_registry = cls.load_from_db(language_id)
        b = datetime.now()
        if output:
            new_registry.print_load_stats(a, b)

        return b - a

    @classmethod
    def profile_load_from_file(cls, file_name, language_id, output=False):

        a = datetime.now()
        new_registry = cls.load_from_file(file_name, language_id)
        b = datetime.now()
        if output:
            new_registry.print_load_stats(a, b)

        return b - a



# "hermit dave" has a nice repository of word frequencies
# computed for many languages based on movie subtitles
from .language_info import LanguageInfo
from .config import DATA_HERMIT_FOLDER


def path_of_hermit_language_file(language):
    file_name = "{0}/{1}/{1}_50k.txt".format(DATA_HERMIT_FOLDER, language)
    return file_name


def load_language_from_hermit(language, hermit_root_folder=None):

    # by default use the hermit folder in the config file
    if not hermit_root_folder:
        hermit_root_folder = DATA_HERMIT_FOLDER

    file_name = "{0}/{1}/{1}_50k.txt".format(hermit_root_folder, language)
    d = LanguageInfo.load_from_file(file_name, language)
    return d


def load_multiple_languages_from_hermit(languages, hermit_root_folder=None):

    # by default use the hermit folder in the config file
    if not hermit_root_folder:
        hermit_root_folder = DATA_HERMIT_FOLDER

    result = dict()
    for language in languages:
        result[language] = load_language_from_hermit(language, hermit_root_folder)

    return result



import math


def compute_frequency(occurrence_count):
    """

        frequency is too coarse.
         taking the log from it is going
         to smooth it somehow

    :param occurrence_count:
    :return:
    """
    return round(math.log(occurrence_count), 2)


def compute_difficulty(rank):
    """

        normalized between 0 and 1
         the difficulty increases in steps of 0.01
         for every 500 ranks
         assumes a frequency list of 50K words

    :param rank:
    :return:
    """
    return rank // 500 / 100.0


def compute_importance(occurrence_count):
    """

        need a better way to think about
        this... but for now we go with
        log of occurrence count

    :param occurrence_count:
    :return:
    """
    return round(math.log(occurrence_count), 2)


def compute_klevel(rank):
    """

        assumes a frequency list of 50K words

    :param rank:
    :return:
    """
    return (rank // 1000) + 1


# Lazy initialized object
from wordstats import LanguageInfo
from .loading_from_hermit import load_language_from_hermit


class Word(object):
    """

    A simple interface for lazily loading individual language info
    when retrieving word info for individual words

    """

    stats_dict = dict()

    @classmethod
    def stats(cls, word, language):
        """

            Assumes that there is information about the given language
            in the wordstats data folder. If not, it will throw an exception.

        :param word: string
        :param language: string
        :return:

            A WordInfo (or an UnknownWordInfo if the word is not
            found in the frequency data)

        """

        if language not in cls.stats_dict:
            cls.stats_dict[language] = LanguageInfo.load(language)

        return cls.stats_dict[language][word]



all_languages = ["da", "de", "el", "en", "es", "fr", "it", "nl", "no", "pl", "pt", "ro", "zh-CN"]

all_languages_with_latin_characters = ["da", "de", "en", "es", "fr", "it", "nl", "no", "pl", "pt", "ro"]



from unittest import TestCase

from wordstats import LanguageInfo
from wordstats.loading_from_hermit import path_of_hermit_language_file
from wordstats.tests.test_basics import SimpleTests


class PerformanceTests(TestCase):

    @classmethod
    def test_compare_performance(self):
        SimpleTests.test_caching_to_db()
        output = False
        time_db = LanguageInfo.profile_load_from_db("de", output=output)
        time_file = LanguageInfo.profile_load_from_file(path_of_hermit_language_file("de"), "de", output=output)
        assert time_db < time_file



from sys import stdout

import sqlalchemy.orm
import sqlalchemy
from sqlalchemy import Column, Integer, String, UniqueConstraint
from sqlalchemy import Float

from .base_service import Base, SimplifiedQuery


class UnknownWordInfo(object):
    """ Default values if  we have no information about
    a given word """
    def __init__(self):
        self.frequency = 0
        self.importance = 0
        self.difficulty = 1
        self.klevel = 100
        self.rank = 100000


class WordInfo(SimplifiedQuery, Base):
    __tablename__ = 'word_info'
    __table_args__ = {'mysql_collate': 'utf8_bin'}

    id = Column(Integer, primary_key=True)

    word = Column(String(255), nullable =False, index = True)
    language_id = Column(String(2), nullable =False, index = True)

    frequency = Column(Integer)
    importance = Column(Integer)
    difficulty = Column(Float)
    rank = Column(Integer)
    klevel = Column(Integer)

    UniqueConstraint(word, language_id)

    def __init__(self, word, language_id, frequency, difficulty, importance, rank, klevel):
        self.word = word
        self.language_id = language_id
        self.frequency = frequency
        self.importance = importance
        self.difficulty = difficulty
        self.klevel = klevel
        self.rank = rank

    def __str__(self):
        result = "{0}: (lang: {1}, rank: {5}, freq: {2}, imp: {3}, diff: {4}, klevel: {6})".format(
            self.word,
            self.language_id,
            self.frequency,
            self.importance,
            self.difficulty,
            self.rank,
            self.klevel)

        return result

    @classmethod
    def find(cls, word, language_id):
        word = word.lower()
        try:
            return (cls.query().filter(cls.word == word)
                    .filter(cls.language_id == language_id)
                    .one())
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    @classmethod
    def find_all(cls,language_id):
        return cls.query().filter(cls.language_id == language_id).all()





from wordstats.config import DATA_COMMON_FOLDER

COMMON_WORDS = []

common_words_file_name = DATA_COMMON_FOLDER + "/common.txt"


def common_words():
    global COMMON_WORDS

    if COMMON_WORDS:
        return COMMON_WORDS

    with open(common_words_file_name) as common_words_file:
        words_list = common_words_file.read().splitlines()
        COMMON_WORDS = words_list
        return COMMON_WORDS


def write_common_words_file():
    """
    Compute common words between all_languages_with_latin_characters (i.e.
        "da","de","en","es","fr","it","nl","no","pl","pt","ro")
    Write them to file
    :return:
    """
    from wordstats import LanguageInfo
    from wordstats.language_codes import all_languages_with_latin_characters

    common_words = set()

    reference_language = LanguageInfo.load(all_languages_with_latin_characters[0])
    all_other_languages = [LanguageInfo.load(each) for each in all_languages_with_latin_characters[1:]]

    with open(common_words_file_name, 'w') as common_words_file:

        for each in reference_language.all_words():
            at_least_one_exception = False
            for other_language in all_other_languages:
                if each not in other_language.all_words():
                    at_least_one_exception = True
                    continue

            if not at_least_one_exception:
                if each not in common_words:
                    common_words.add(each)
                    common_words_file.write(each + "\n")
                    print(each)



from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from .config import db_uri

# This thing will be the superclass of all our model classes
Base = declarative_base()


# This is where we'll be using the session from
class BaseService(object):
    # assumes the existence of a ./wordranks/config.cfg

    engine = create_engine(db_uri, encoding='utf-8')
    Session = sessionmaker(bind=engine)
    session = Session()

    @classmethod
    def drop_tables(cls):
        # We have to do a commit() before the drop_all()
        # Otherwise the system just freezes sometimes!
        cls.session.commit()
        cls.session.close_all()

        # Initial cleanup
        Base.metadata.reflect(cls.engine)
        Base.metadata.drop_all(cls.engine)
        # Creating the tables again
        Base.metadata.create_all(cls.engine)



class SimplifiedQuery(object):

    @classmethod
    def query(cls):
        return BaseService.session.query(cls)


# in this file, we must first install the DB
# load the words in the DB
# then the user can simply import the lib and start
# using it


# !/usr/bin/env python
# -*- coding: utf8 -*-
import os

import setuptools


def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            if filename.endswith(".txt"):
                paths.append(os.path.join('..', path, filename))
    return paths


extra_files = package_files('wordstats/language_data/')

with open('README.md') as f:
    long_description = f.read()

setuptools.setup(
    name="wordstats",
    packages=setuptools.find_packages(),
    version="1.0.3",
    license="MIT",
    description="Multilingual word frequency statistics for Python based on subtitles corpora",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Mircea Lungu",
    author_email="me@mir.lu",
    url="https://github.com/zeeguu-ecosystem/Python-Wordstats",
    download_url="https://github.com/zeeguu-ecosystem/Python-Wordstats/archive/v_1.0.3.tar.gz",
    include_package_data=True,
    zip_safe=False,
    keywords="natural language processing, multilingual",
    package_data={'': extra_files},
    install_requires=("configobj",
                      "sqlalchemy"),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Text Processing :: Linguistic',
        'License :: OSI Approved :: MIT License',  # Again, pick a license
        'Programming Language :: Python :: 3',  # Specify which pyhton versions that you want to support
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

)



import os

import platform
import tempfile
import logging as log

tempdir = "/tmp" if platform.system() == "Darwin" else tempfile.gettempdir()

db_uri = 'sqlite:///' + tempdir  + '/wordinfo.db'

log.debug("running with DB URI: " + db_uri )

# for mysql, we want to declare the default character encoding
# for comm. with the db
if db_uri.startswith("mysql"):
    db_uri += '?charset=utf8'

DATA_HERMIT_FOLDER = 'language_data/hermitdave/2016'

package_directory = os.path.dirname(os.path.abspath(__file__))
DATA_COMMON_FOLDER = package_directory + os.sep +'language_data/common'

MAX_WORDS = 10000
